"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.deployDataProtobufSerialize = exports.verifyDeploy = exports.signDeploy = void 0;
var blakejs_1 = require("blakejs");
var elliptic_1 = require("elliptic");
var google_protobuf_1 = __importDefault(require("google-protobuf"));
/**
  * Signs deploy data.
  *
  * The private key for signing can be in different formats supported by
  * [elliptic](https://github.com/indutny/elliptic#ecdsa) library.
  *
  * **NOTE: Signing function can be used independently without this library and JS generated code (see _rnode-sign.ts_ source).**
  *
  * ```typescript
  * // Generate new key pair
  * const { ec } = require('elliptic')
  * const secp256k1 = new ec('secp256k1')
  * const key = secp256k1.genKeyPair()
  *
  * // Or use existing private key as hex string, Uint8Array, Buffer or ec.KeyPair
  * const key = '1bf36a3d89c27ddef7955684b97667c75454317d8964528e57b2308947b250b0'
  *
  * const deployData = {
  *   term: 'new out(`rho:io:stdout`) in { out!("Browser deploy test") }',
  *   timestamp: Date.now(),      // nonce
  *   phloprice: 1,               // price of tinny REV for phlogiston (gas)
  *   phlolimit: 10e3,            // max phlogiston (gas) for deploy execution
  *   validafterblocknumber: 123, // latest block number
  *   shardid: 'testnet',         // shard name
  * }
  *
  * // Signed deploy with deployer, sig and sigalgorithm fields populated
  * const signed = signDeploy(key, deployData)
  * ```
  */
var signDeploy = function (privateKey, deployObj) {
    var term = deployObj.term, timestamp = deployObj.timestamp, phlolimit = deployObj.phlolimit, phloprice = deployObj.phloprice, validafterblocknumber = deployObj.validafterblocknumber, shardid = deployObj.shardid;
    // Currently supported algorithm
    var sigalgorithm = 'secp256k1';
    // Serialize deploy data for signing
    var deploySerialized = (0, exports.deployDataProtobufSerialize)({
        term: term,
        timestamp: timestamp,
        phlolimit: phlolimit,
        phloprice: phloprice,
        validafterblocknumber: validafterblocknumber,
        shardid: shardid,
    });
    // Signing key
    var crypt = new elliptic_1.ec(sigalgorithm);
    var key = getSignKey(crypt, privateKey);
    var deployer = Uint8Array.from(key.getPublic('array'));
    // Hash and sign serialized deploy
    var hashed = (0, blakejs_1.blake2bHex)(deploySerialized, void 666, 32);
    var sigArray = key.sign(hashed, { canonical: true }).toDER();
    var sig = Uint8Array.from(sigArray);
    // Return deploy object / ready for sending to RNode
    return {
        term: term,
        timestamp: timestamp,
        phlolimit: phlolimit,
        phloprice: phloprice,
        validafterblocknumber: validafterblocknumber,
        shardid: shardid,
        sigalgorithm: sigalgorithm,
        deployer: deployer,
        sig: sig,
    };
};
exports.signDeploy = signDeploy;
/**
  * Verifies deploy for a valid signature.
  */
var verifyDeploy = function (deployObj) {
    var term = deployObj.term, timestamp = deployObj.timestamp, phlolimit = deployObj.phlolimit, phloprice = deployObj.phloprice, validafterblocknumber = deployObj.validafterblocknumber, shardid = deployObj.shardid, sigalgorithm = deployObj.sigalgorithm, deployer = deployObj.deployer, sig = deployObj.sig;
    // Serialize deploy data for signing
    var deploySerialized = (0, exports.deployDataProtobufSerialize)({
        term: term,
        timestamp: timestamp,
        phlolimit: phlolimit,
        phloprice: phloprice,
        validafterblocknumber: validafterblocknumber,
        shardid: shardid,
    });
    // Signing public key to verify
    var crypt = new elliptic_1.ec(sigalgorithm);
    var key = crypt.keyFromPublic(deployer);
    // Hash and verify signature
    var hashed = (0, blakejs_1.blake2bHex)(deploySerialized, void 666, 32);
    var isValid = key.verify(hashed, sig);
    return isValid;
};
exports.verifyDeploy = verifyDeploy;
/**
 * Serialization of DeployDataProto object without generated JS code.
 */
var deployDataProtobufSerialize = function (deployData) {
    var term = deployData.term, timestamp = deployData.timestamp, phlolimit = deployData.phlolimit, phloprice = deployData.phloprice, validafterblocknumber = deployData.validafterblocknumber, shardid = deployData.shardid;
    // Create binary stream writer
    var writer = new google_protobuf_1.default.BinaryWriter();
    // Write fields (protobuf doesn't serialize default values)
    var writeString = function (order, val) { return val != "" && writer.writeString(order, val); };
    var writeInt64 = function (order, val) { return val != 0 && writer.writeInt64(order, val); };
    // https://github.com/rchain/rchain/blob/ebe4d476371/models/src/main/protobuf/CasperMessage.proto#L134-L149
    // message DeployDataProto {
    //   bytes  deployer             = 1; //public key
    //   string term                 = 2; //rholang source code to deploy (will be parsed into `Par`)
    //   int64  timestamp            = 3; //millisecond timestamp
    //   bytes  sig                  = 4; //signature of (hash(term) + timestamp) using private key
    //   string sigAlgorithm         = 5; //name of the algorithm used to sign
    //   int64 phloPrice             = 7; //phlo price
    //   int64 phloLimit             = 8; //phlo limit for the deployment
    //   int64 validAfterBlockNumber = 10;
    //   string shardId              = 11;//shard ID to prevent replay of deploys between shards
    // }
    // Serialize fields
    writeString(2, term);
    writeInt64(3, timestamp);
    writeInt64(7, phloprice);
    writeInt64(8, phlolimit);
    writeInt64(10, validafterblocknumber);
    writeString(11, shardid);
    return writer.getResultBuffer();
};
exports.deployDataProtobufSerialize = deployDataProtobufSerialize;
/**
 * Fix for ec.keyFromPrivate not accepting KeyPair.
 * - detect KeyPair if it have `sign` function
 */
var getSignKey = function (crypt, pk) {
    return pk && typeof pk != 'string' && pk.sign && pk.sign.constructor == Function ? pk : crypt.keyFromPrivate(pk);
};
//# sourceMappingURL=rnode-sign.js.map